# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
from similarities.utils.get_file import http_get
from similarities.utils.util import cos_sim, dot_score, pairwise_dot_score, pairwise_cos_sim, normalize_embeddings, \
    semantic_search, paraphrase_mining_embeddings, community_detection
